/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

namespace fe
{

void CurveCreateOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("interpolation")="power";
	catalog<String>("interpolation","label")="Interpolation";
	catalog<String>("interpolation","choice:0")="linear";
	catalog<String>("interpolation","label:0")="Linear";
	catalog<String>("interpolation","choice:1")="pointNormal";
	catalog<String>("interpolation","label:1")="Point Normal";
	catalog<String>("interpolation","choice:2")="cardinal";
	catalog<String>("interpolation","label:2")="Cardinal Spline";
	catalog<String>("interpolation","choice:3")="power";
	catalog<String>("interpolation","label:3")="Power Bisect";

	catalog<String>("method")="length";
	catalog<String>("method","label")="Method";
	catalog<String>("method","choice:0")="length";
	catalog<String>("method","label:0")="Length";
	catalog<String>("method","choice:1")="count";
	catalog<String>("method","label:1")="Count";

	catalog<Real>("length")=1.0;
	catalog<Real>("length","low")=0.0;
	catalog<Real>("length","high")=1.0;
	catalog<Real>("length","min")=0.0;
	catalog<Real>("length","max")=1e3;
	catalog<String>("length","label")="Segment Length";

	catalog<I32>("segments")=4;
	catalog<I32>("segments","low")=1;
	catalog<I32>("segments","high")=100;
	catalog<I32>("segments","min")=1;
	catalog<I32>("segments","max")=1000000;
	catalog<String>("segments","label")="Segments";

	catalog< sp<Component> >("Input Surface");
	catalog<bool>("Input Surface","optional")=true;

	//* no auto-copy
	catalog< sp<Component> >("Output Curve");
	catalog<String>("Output Curve","IO")="output";
}

void CurveCreateOp::handle(Record& a_rSignal)
{
	const String interpolation=catalog<String>("interpolation");
	const String method=catalog<String>("method");
	const Real length=catalog<Real>("length");
	const I32 segments=catalog<I32>("segments");

	const BWORD doLinear=(interpolation=="linear");
	const BWORD doCardinal=(interpolation=="cardinal");
	const BWORD doPointNormal=(interpolation=="pointNormal");
	const BWORD doPower=(interpolation=="power");

	sp<DrawI> spDrawI;
	if(!accessDraw(spDrawI,a_rSignal)) return;

	sp<DrawI> spDrawGuide;
//	accessGuide(spDrawGuide,a_rSignal,e_warning);

	sp<SurfaceAccessorI> spInputPosition;
	access(spInputPosition,"Input Surface",
			e_point,e_position,e_quiet);

	sp<SurfaceAccessorI> spInputNormal;
	if(doPointNormal || doPower)
	{
		access(spInputNormal,"Input Surface",
				e_point,e_normal,e_warning);
	}

	I32 pointCount=0;

	if(spInputPosition.isValid())
	{
		pointCount=spInputPosition->count();
	}

	if(pointCount<2)
	{
		//* no input path

		SpatialVector pointArray[segments+1];

		for(I32 index=0;index<=segments;index++)
		{
			set(pointArray[index],0.001*index*index,index*length);
		}

		spDrawI->drawLines(pointArray,NULL,segments+1,
				DrawI::e_strip,false,NULL);
	}
	else
	{
		MatrixSqrt<SpatialTransform> matrixSqrt;
		MatrixPower<SpatialTransform> matrixPower;

		Real totalDistance=0.0;
		Real distanceArray[pointCount];

		SpatialVector point0=spInputPosition->spatialVector(0);
		distanceArray[0]=0.0;

		for(I32 pointIndex=1;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector point1=
					spInputPosition->spatialVector(pointIndex);

			totalDistance+=magnitude(point1-point0);
			distanceArray[pointIndex]=totalDistance;

			point0=point1;
		}

		const I32 trueSegments=(method=="length" && length>0.0)?
				totalDistance/length: segments;

		SpatialVector pointArray[trueSegments+1];

		I32 cv2=0;

		for(I32 index=0;index<=trueSegments;index++)
		{
			const Real along=index/Real(trueSegments);

			const Real distance=along*totalDistance;

			while(distanceArray[cv2]<=distance && cv2<pointCount-1)
			{
				cv2++;
			}

			const I32 cv1=fe::maximum(cv2-1,0);

			const Real t=(distance-distanceArray[cv1])/
					(distanceArray[cv2]-distanceArray[cv1]);

			const SpatialVector p1=spInputPosition->spatialVector(cv1);
			const SpatialVector p2=spInputPosition->spatialVector(cv2);

			if(doLinear)
			{
				pointArray[index]=p1*(Real(1)-t)+p2*t;
				continue;
			}

			if(doPointNormal)
			{
				SpatialVector n1(0.0,1.0,0.0);
				SpatialVector n2(0.0,1.0,0.0);

				if(spInputNormal.isValid())
				{
					n1=spInputNormal->spatialVector(cv1);
					n2=spInputNormal->spatialVector(cv2);
				}

				TrianglePN<Real> trianglePN;
				trianglePN.configure(p1,p2,p2,n1,n2,n2);

				const SpatialBary bary(Real(1)-t,t,Real(0));
				SpatialVector point;
				SpatialVector norm;
				trianglePN.solve(bary,point,norm);

				pointArray[index]=point;
				continue;
			}

			const I32 cv0=fe::maximum(cv2-2,0);
			const I32 cv3=fe::minimum(cv2+1,pointCount-1);

			const SpatialVector p0=spInputPosition->spatialVector(cv0);
			const SpatialVector p3=spInputPosition->spatialVector(cv3);

			if(doCardinal)
			{
				pointArray[index]=
						Spline<SpatialVector,Real>::Cardinal2D(t,p0,p1,p2,p3);
				continue;
			}

			const SpatialVector t1=unitSafe(p2-p0);
			const SpatialVector t2=unitSafe(p3-p1);

			SpatialVector n1(0.0,1.0,0.0);
			SpatialVector n2(0.0,1.0,0.0);

			if(spInputNormal.isValid())
			{
				n1=spInputNormal->spatialVector(cv1);
				n2=spInputNormal->spatialVector(cv2);
			}

			SpatialTransform xform1;
			SpatialTransform xform2;
			makeFrameTangentX(xform1,p1,t1,n1);
			makeFrameTangentX(xform2,p2,t2,n2);

			SpatialTransform inv1;
			invert(inv1,xform1);

			SpatialTransform delta=inv1*xform2;

			SpatialTransform half;
			matrixSqrt.solve(half,delta);

			const SpatialTransform halfway=xform1*half;

			const SpatialVector pMid=halfway.translation();
			const SpatialVector tMid=unitSafe(p2-p1);
			const SpatialVector nMid=unitSafe(n1+n2);

			SpatialTransform middle;
			makeFrameTangentX(middle,pMid,tMid,nMid);

			SpatialTransform xformA;
			SpatialTransform xformB;
			Real tt;

			SpatialTransform invA;

			if(t<0.5)
			{
				xformA=xform1;
				xformB=middle;
				tt=t*Real(2);
				invA=inv1;
			}
			else
			{
				xformA=middle;
				xformB=xform2;
				tt=t*Real(2)-Real(1);
				invert(invA,xformA);
			}

			SpatialTransform subDelta=invA*xformB;

			SpatialTransform partial;
			matrixPower.solve(partial,subDelta,tt);

			SpatialTransform between=xformA*partial;

//			feLog(">>>> %d/%d t %.6G cv %d %d %d %d\n",
//					index,trueSegments,t,cv0,cv1,cv2,cv3);
//			feLog("delta\n%s\n",c_print(delta));
//			feLog("halfway\n%s\n",c_print(halfway));
//			feLog("middle\n%s\n",c_print(middle));
//			feLog("xformA\n%s\n",c_print(xformA));
//			feLog("xformB\n%s\n",c_print(xformB));
//			feLog("subDelta\n%s\n",c_print(subDelta));
//			feLog("partial\n%s\n",c_print(partial));

			pointArray[index]=between.translation();

			if(spDrawGuide.isValid())
			{
				const Color lightred(1.0,0.3,0.3,1.0);
				const Color lightgreen(0.3,1.0,0.3,1.0);
				const Color lightblue(0.3,0.3,1.0,1.0);

				SpatialVector line[2];

				for(U32 m=0;m<4;m++)
				{
					const SpatialTransform step=
							m==0? xform1:
							(m==1? xform2:
							(m==2? middle: between));

					const Real scalar=
							m<2? 0.1: (m==2? 0.05: 0.03);

					line[0]=step.translation();

					for(U32 n=0;n<3;n++)
					{
						const Color color=
								n==0? lightred: (n==1? lightgreen: lightblue);

						line[1]=line[0]+scalar*step.column(n);

						spDrawGuide->drawLines(line,NULL,2,
								DrawI::e_strip,false,&color);
					}
				}
			}
		}

		spDrawI->drawLines(pointArray,NULL,trueSegments+1,
				DrawI::e_strip,false,NULL);
	}
}

} // namespace
