/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "oplab/oplab.pmh"

using namespace fe;

void RecordOp::Handler::handle(Record &a_signal)
{
	safeLock();

	feLog("RecordOp::Handler::handle ttl %d\n",m_hpRecordOp->m_ttl(a_signal));

	if(m_hpRecordOp->m_source.check(a_signal))
	{
		const String source=m_hpRecordOp->m_source(a_signal);

		feLog("  source \"%s\" vs \"%s\"\n",
				source.c_str(),m_self.c_str());

		if(source==m_self)
		{
			safeUnlock();
			return;
		}
	}
	if(m_hpRecordOp->m_text.check(a_signal))
	{
		m_state=m_hpRecordOp->m_text(a_signal).c_str();

		feLog("  text \"%s\"\n",m_state.c_str());
	}
	if(m_hpRecordOp->m_model.check(a_signal))
	{
		m_record=m_hpRecordOp->m_model(a_signal);
		m_state="changed";
		m_hpRecordOp->dirty();

		feLog("  model %d\n",m_record.isValid());
	}

	safeUnlock();
}

String RecordOp::Handler::state(void) const
{
	safeLock();
	String result=m_state;
	safeUnlock();

	return result;
}

void RecordOp::Handler::setState(String a_state)
{
	safeLock();
	m_state=a_state;
	safeUnlock();
}

Record RecordOp::Handler::record(void) const
{
	safeLock();
	Record result=m_record;
	safeUnlock();

	return result;
}

void RecordOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	//* page File
	catalog<String>("savefile")="";
	catalog<String>("savefile","label")="Save File";
	catalog<String>("savefile","suggest")="filename";
	catalog<String>("savefile","page")="File";

	catalog<String>("loadfile")="";
	catalog<String>("loadfile","label")="Load File";
	catalog<String>("loadfile","suggest")="filename";
	catalog<String>("loadfile","page")="File";

	//* page Net
	catalog<String>("role")="None";
	catalog<String>("role","label")="Role";
	catalog<String>("role","choice:0")="None";
	catalog<String>("role","choice:1")="Server";
	catalog<String>("role","choice:2")="Client";
	catalog<String>("role","page")="Net";

	catalog<String>("address")="127.0.0.1";
	catalog<String>("address","label")="Address";
	catalog<String>("address","page")="Net";

	catalog<I32>("port")=7876;
	catalog<I32>("port","max")=65535;
	catalog<String>("port","label")="Port";
	catalog<String>("port","page")="Net";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","prompt")="watching for messages";
	catalog<String>("Brush","idle")="any";

	catalog< sp<Component> >("Input Surface");
	catalog<bool>("Input Surface","optional")=true;

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
//	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;

	m_spScope=registry()->create("Scope");

	m_ttl.setup(m_spScope,FE_USE(":TTL"));
	m_source.setup(m_spScope,"source");
	m_text.setup(m_spScope,"text");
	m_model.setup(m_spScope,"model");

	m_spLayoutHB=m_spScope->declare("heartbeat");

	m_spLayoutMessage=m_spScope->declare("message");
	m_spLayoutMessage->populate(m_ttl);
	m_spLayoutMessage->populate(m_source);
	m_spLayoutMessage->populate(m_text);

	m_spLayoutSurface=m_spScope->declare("surface");
	m_spLayoutSurface->populate(m_ttl);
	m_spLayoutSurface->populate(m_source);
	m_spLayoutSurface->populate(m_model);
}

void RecordOp::handle(Record& a_rSignal)
{
	const String savefile=catalog<String>("savefile","eval");
	const String loadfile=catalog<String>("loadfile","eval");
	const String role=catalog<String>("role");
	const String address=catalog<String>("address");
	const I32 port=catalog<I32>("port");

	m_spScope->setName(name()+".Scope");

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	if(spDrawBrush.isValid())
	{
		m_brushed=TRUE;

		if(role=="Client")
		{
			Record heartbeat=m_spScope->createRecord(m_spLayoutHB);
			m_spSignaler->signal(heartbeat);
		}
		return;
	}

	const Real frame=currentFrame(a_rSignal);
	const BWORD frameChanged=(frame!= m_lastFrame);
	const BWORD replaced=(catalog<bool>("Output Surface","replaced"));
	const BWORD paramChanged=(!m_brushed && !replaced && !frameChanged);
	m_brushed=FALSE;
	m_lastFrame=frame;

	const BWORD rebuild=(paramChanged || frameChanged || replaced);

	feLog("RecordOp::handle frame %.6G changed %d replaced %d"
			" param %d rebuild %d brushed %d\n",
			frame,frameChanged,replaced,paramChanged,rebuild,m_brushed);

	sp<SurfaceAccessibleI> spInputAccessible;
	sp<SurfaceAccessibleBase> spInputBase;

	if(!savefile.empty() || role=="Server")
	{
		if(!access(spInputAccessible,"Input Surface")) return;

		spInputBase=registry()->create("*.SurfaceAccessibleBase");
		if(spInputBase.isValid())
		{
			spInputBase->bind(m_spScope);

			feLog("RecordOp::handle convert input\n");
			spInputBase->copy(spInputAccessible);
		}
	}

	sp<SurfaceAccessibleBase> spOutputBase;

	if(!savefile.empty() && spInputBase.isValid())
	{
		feLog("RecordOp::handle save \"%s\"\n",savefile.c_str());
		spInputBase->save(savefile);
	}

	if(m_spServer.isValid() && role!="Server")
	{
		m_spServer=NULL;
	}
	if(m_spClient.isValid() && role!="Client")
	{
		m_spClient=NULL;
	}

	if(role=="None")
	{
		m_spHandler=NULL;
		m_spSignaler=NULL;
	}
	else
	{
		Record heartbeat=m_spScope->createRecord(m_spLayoutHB);

		Record signalMessage=m_spScope->createRecord(m_spLayoutMessage);
		m_source(signalMessage)=role;

		if(m_spHandler.isNull())
		{
			m_spHandler=new Handler;
			m_spHandler->bind(sp<RecordOp>(this));
			m_spHandler->setName(name()+".Handler");
			m_spHandler->setSelf(role);
		}

		if(m_spSignaler.isNull())
		{
			m_spSignaler=registry()->create("SignalerI");
			m_spSignaler->insert(m_spHandler,m_spLayoutMessage);
			m_spSignaler->insert(m_spHandler,m_spLayoutSurface);
		}
		m_spSignaler->setName(name()+".Signaler");

		if(role=="Server" && spInputBase.isValid())
		{
			if(m_spServer.isNull())
			{
				m_spServer=registry()->create("ServerI");
				if(!m_spServer.isValid())
				{
					feX("RecordOp::handle","could not create ServerI");
				}

				m_spServer->addLayout(m_spLayoutMessage);
				m_spServer->addLayout(m_spLayoutSurface);

				m_spServer->start(m_spSignaler,(short)port,m_spLayoutHB);
			}

			Record surfaceRecord=spInputBase->record();

			Record signalSurface=m_spScope->createRecord(m_spLayoutSurface);
			m_ttl(signalSurface)=2;
			m_source(signalSurface)=role;
			m_model(signalSurface)=surfaceRecord;

			feLog("RecordOp::handle server signal HB; wait for 'ready'\n");
			while(m_spHandler->state()!="ready")
			{
				m_spSignaler->signal(heartbeat);
				milliSleep(100);
			}

			feLog("RecordOp::handle server signal surface\n");
			m_spSignaler->signal(signalSurface);
			m_spSignaler->signal(heartbeat);
		}

		if(role=="Client")
		{
			if(m_spClient.isNull())
			{
				m_spClient=registry()->create("ClientI");
				if(!m_spClient.isValid())
				{
					feX("RecordOp::handle","could not create ClientI");
				}

				m_spClient->addLayout(m_spLayoutMessage);
				m_spClient->addLayout(m_spLayoutSurface);
//				m_spClient->setScope(m_spScope);

				m_spClient->start(m_spSignaler,
						address,(short)port,m_spLayoutHB);
			}

			m_text(signalMessage)="ready";
			m_ttl(signalMessage)=2;

			feLog("RecordOp::handle client signal 'ready'\n");
			m_spSignaler->signal(signalMessage);
			m_spSignaler->signal(heartbeat);

			feLog("RecordOp::handle client check for 'changed'\n");
			if(rebuild || m_spHandler->state()=="changed")
			{
				m_spHandler->setState("current");

				Record surfaceRecord=m_spHandler->record();
				if(surfaceRecord.isValid())
				{
					spOutputBase=registry()->create("*.SurfaceAccessibleBase");
					spOutputBase->bind(m_spScope);
					spOutputBase->bind(surfaceRecord);
				}
			}
		}
	}

	if(!loadfile.empty())
	{
		feLog("RecordOp::handle load \"%s\"\n",loadfile.c_str());

		spOutputBase=registry()->create("*.SurfaceAccessibleBase");
		spOutputBase->bind(m_spScope);
		if(!spOutputBase->load(loadfile))
		{
			spOutputBase=NULL;
		}
	}

	if(spOutputBase.isValid())
	{
		feLog("RecordOp::handle convert output\n");

		sp<SurfaceAccessibleI> spOutputAccessible;
		if(!accessOutput(spOutputAccessible,a_rSignal)) return;

		spOutputAccessible->copy(spOutputBase);
	}
}
